import java.util.*;
import java.util.function.BiFunction;
import java.util.stream.Collectors;

public class DictionaryTree implements Comparable<DictionaryTree> {
    private int popularity;
    private Map<Character, DictionaryTree> children = new LinkedHashMap<>();
    private String character;
    private DictionaryTree parent;

    public DictionaryTree() {
        this.popularity = -1;
        this.character = "";
        this.parent = null;
    }

    public DictionaryTree(char c, DictionaryTree parent) {
        this.popularity = -1;
        this.character = c + "";
        this.parent = parent;
    }

    /**
     * Inserts the given word into this dictionary.
     * If the word already exists, nothing will change.
     *
     * @param word the word to insert
     */
    void insert(String word) {
        insert(word, 0);
    }

    //Adds a new word to the trie tree.
    private void insertWord(String string, DictionaryTree dictionaryTree, int popularity) {
        final DictionaryTree nextChild;
        if (dictionaryTree.children.containsKey(string.charAt(0))) {
            nextChild = dictionaryTree.children.get(string.charAt(0));
        } else {
            nextChild = new DictionaryTree(string.charAt(0), dictionaryTree);
            dictionaryTree.children.put(string.charAt(0), nextChild);
        }

        if (string.length() == 1) {
            nextChild.popularity = popularity;
        } else {
            insertWord(string.substring(1), nextChild, popularity);
        }
    }

    /**
     * Inserts the given word into this dictionary with the given popularity.
     * If the word already exists, the popularity will be overriden by the given value.
     *
     * @param word       the word to insert
     * @param popularity the popularity of the inserted word
     */
    void insert(String word, int popularity) {
        if (!children.containsKey(word.charAt(0))) {
            children.put(word.charAt(0), new DictionaryTree(word.charAt(0), null));
        }
        insertWord(word.substring(1), children.get(word.charAt(0)), popularity);
    }

    /**
     * Removes the specified word from this dictionary.
     * Returns true if the caller can delete this node without losing
     * part of the dictionary, i.e. if this node has no children after
     * deleting the specified word.
     *
     * @param word the word to delete from this dictionary
     * @return whether or not the parent can delete this node from its children
     */
    boolean remove(String word) {
        if (word.length() > 0) {
            if (children.containsKey(word.charAt(0))) {
                DictionaryTree dictionaryTree = children.get(word.charAt(0));
                if (word.length() == 1) {
                    if (dictionaryTree.children.size() > 0) {
                        children.remove(word.charAt(0));
                        return true;
                    } else {
                        children.remove(word.charAt(0));
                        return false;
                    }
                } else {
                    return dictionaryTree.remove(word.substring(1));
                }
            }
        }
        return true;
    }

    /**
     * Determines whether or not the specified word is in this dictionary.
     *
     * @param word the word whose presence will be checked
     * @return true if the specified word is stored in this tree; false otherwise
     */
    boolean contains(String word) {
        if (word.length() == 0) {
            return popularity > -1;
        }

        return children.containsKey(word.charAt(0)) && children.get(word.charAt(0)).contains(word.substring(1));
    }

    /**
     * @param prefix the prefix of the word returned
     * @return a word that starts with the given prefix, or an empty optional
     * if no such word is found.
     */
    Optional<String> predict(String prefix) {
        List<String> strings = predict(prefix, 1);
        if (strings.size() > 0) {
            return Optional.of(strings.get(0));
        } else {
            return Optional.empty();
        }
    }

    /**
     * Predicts the (at most) n most popular full English words based on the specified prefix.
     * If no word with the specified prefix is found, an empty Optional is returned.
     *
     * @param prefix the prefix of the words found
     * @return the (at most) n most popular words with the specified prefix
     */
    List<String> predict(String prefix, int n) {
        if (prefix.length() == 0) {
            List<DictionaryTree> nodes = this.allNodes()
                    .stream().filter(dictionaryTree -> dictionaryTree.popularity > -1).sorted().collect(Collectors.toList());
            List<DictionaryTree> dictionaryTrees = nodes.subList(0, n);
            List<String> result = new ArrayList<>();
            for (DictionaryTree dictionaryTree :
                    dictionaryTrees) {
                result.add(dictionaryTree.getWord());
            }
            return result;
        }

        if (children.containsKey(prefix.charAt(0))) {
            return children.get(prefix.charAt(0)).predict(prefix.substring(1), n);
        }
        return new ArrayList<>();
    }

    private String getWord() {
        StringBuilder wordBuilder = new StringBuilder();
        DictionaryTree node = this;

        while (node != null) {
            wordBuilder.append(node.character);
            node = node.parent;
        }
        return wordBuilder.reverse().toString();
    }

    private List<DictionaryTree> allNodes() {
        List<DictionaryTree> listOfNodes = new ArrayList<>();
        addAllNodes(this, listOfNodes);
        return listOfNodes;
    }

    private static void addAllNodes(DictionaryTree node, List<DictionaryTree> listOfNodes) {
        if (node != null) {
            listOfNodes.add(node);
            Collection<DictionaryTree> children = node.children.values();
            for (DictionaryTree child : children) {
                addAllNodes(child, listOfNodes);
            }
        }
    }

    /**
     * @return the number of leaves in this tree, i.e. the number of words which are
     * not prefixes of any other word.
     */
    int numLeaves() {
        int numLeaves = 1;
        for (Map.Entry<Character, DictionaryTree> entry : children.entrySet()) {
            numLeaves += entry.getValue().numLeaves();
        }
        return numLeaves;
    }

    /**
     * @return the maximum number of children held by any node in this tree
     */
    int maximumBranching() {
        int maximumBranching = children.size();
        for (Map.Entry<Character, DictionaryTree> entry : children.entrySet()) {
            int numberOfChildren = entry.getValue().maximumBranching();
            maximumBranching = Math.max(maximumBranching, numberOfChildren);
        }
        return maximumBranching;
    }

    /**
     * @return the height of this tree, i.e. the length of the longest branch
     */
    int height() {
        int height = 0;
        for (Map.Entry<Character, DictionaryTree> entry : children.entrySet()) {
            height = Math.max(height, 1 + entry.getValue().height());
        }
        return height;
    }

    /**
     * @return the number of nodes in this tree
     */
    int size() {
        int size = 1;
        for (Map.Entry<Character, DictionaryTree> entry : children.entrySet()) {
            size += entry.getValue().size();
        }
        return size;
    }

    /**
     * @return the longest word in this tree
     */
    String longestWord() {
        String longestWord = this.getWord();
        for (DictionaryTree node : children.values()) {
            String word = node.longestWord();
            if (longestWord.length() < word.length()) {
                longestWord = word;
            }
        }
        return longestWord;
    }

    /**
     * @return all words stored in this tree as a list
     */
    List<String> allWords() {
        List<DictionaryTree> nodes = this.allNodes()
                .stream().filter(dictionaryTree -> dictionaryTree.popularity > -1).sorted().collect(Collectors.toList());
        List<String> result = new ArrayList<>();
        for (DictionaryTree dictionaryTree :
                nodes) {
            result.add(dictionaryTree.getWord());
        }
        return result;
    }

    /**
     * Folds the tree using the given function. Each of this node's
     * children is folded with the same function, and these results
     * are stored in a collection, cResults, say, then the final
     * result is calculated using f.apply(this, cResults).
     *
     * @param f   the summarising function, which is passed the result of invoking the given function
     * @param <A> the type of the folded value
     * @return the result of folding the tree using f
     */
    <A> A fold(BiFunction<DictionaryTree, Collection<A>, A> f) {
        Collection<A> cResults = new ArrayList<>();
        for (DictionaryTree dictionaryTree : children.values()) {
            cResults.add(dictionaryTree.fold(f));
        }
        return f.apply(this, cResults);
    }


    @Override
    public int compareTo(DictionaryTree o) {
        return Integer.compare(o.popularity, popularity);
    }
}
